@@ -64,6 +64,8 @@
  * User-defined bell sound, new setterm control sequences and printk
  * redirection by Martin Mares <mj@k332.feld.cvut.cz> 19-Nov-95
  *
+ * APM screenblank bug fixed Takashi Manabe <manabe@roy.dsl.tutics.tut.jp>
+ * Backported to 2.0.31 by Adam Bradley <artdodge@cs.bu.edu>
  */
 
 #define BLANK 0x0020
@@ -2383,16 +2385,18 @@ void do_blank_screen(int nopowersave)
 #endif /* CONFIG_MULTIMON */
 	console_blanked = fg_console + 1;
 
-#ifdef CONFIG_APM
-	if (apm_display_blank())
-		return;
-#endif
 	if(!nopowersave)
+	{
+#ifdef CONFIG_APM
+		if (apm_display_blank())
+			return;
+#endif
 #ifndef CONFIG_MULTIMON
-	    vesa_blank();
+		vesa_blank();
 #else
-	    vesa_blank(current_monitor);
+		vesa_blank(current_monitor);
 #endif /* CONFIG_MULTIMON */
+	}
 }
 
 void do_unblank_screen(void)
@@ -2707,3 +2711,25 @@ int con_get_font (int mon, char *arg)
 	return set_get_font (mon,arg,0,video_mode_512ch);
 #endif /* CONFIG_MULTIMON */
 }
+
+/*
+ * Report the current status of the vc. This is exported to modules (ARub)
+ */
+
+int con_get_info(int *mode, int *shift, int *col, int *row,
+			struct tty_struct **tty)
+{
+#ifdef CONFIG_MULTIMON
+   	int currcons = fg_console;
+#endif /* CONFIG_MULTIMON */
+	extern int shift_state;
+	extern struct tty_driver console_driver;
+	struct tty_struct **console_table=console_driver.table;
+
+	if (mode) *mode = vt_cons[fg_console]->vc_mode;
+	if (shift) *shift = shift_state;
+	if (col) *col = video_num_columns;
+	if (row) *row = video_num_lines;
+	if (tty) *tty = console_table[fg_console];
+	return fg_console;
+}
